#!/bin/bash

AUTHOR="Janne Hellas"
VERSION="0.4"
LICENSE="Public Domain or equivalent."
LINFO="This software is provided \"as-is\" without any warranties or liabilities."

if [[ $# -eq 0 ]]; then
   echo "FF6 GBA to PC/Mobile save importer/converter v$VERSION"
   echo ""
   echo "Usage: ./ff6mob_gbasave_import.sh [GBA Save file] [Donor PC/Mobile save file]"
   echo ""
   echo "The utility will create a file called 'data_import' from the input files."
   echo "Relevant slot 1 data blocks are replaced by the input GBA save file's slot 1 data."
   echo "Rest of the data is augmented from the donor file. If no donor is specified, an empty one will be utilized."
   echo "Nothing is written to the input files."
   echo ""
   echo "License: $LICENSE"
   echo "$LINFO"
   exit 0
fi

SCRIPTDIR=$(dirname "${BASH_SOURCE}")
DONOR="/tmp/empty_mob$$.sav"
# handle non-option arguments
if [[ $# -gt 2 ]]; then
    echo "$0: Two arguments are required." 1>&2
    exit 1
fi

if [[ ! -f "$1" ]]; then
   echo "$0: Input file doesn't exist." 1>&2
   exit 2
fi

# Detect ff6 advance
FF6ADV=`"${SCRIPTDIR}/util/ff6detect.sh" "$1"`
if [ "${FF6ADV}" != "ff6advance" ]; then
    echo "$0: Input save file format failure." 1>&2
    exit 3
fi

SLOT=1
OUTFILE="data_import"

if [[ "$1" == "$OUTFILE" ]] || [[ "$2" == "$OUTFILE" ]]; then
   echo "$0: Input file cannot be named the same as output." 1>&2
   exit 3
fi

if [[ -f "$OUTFILE" ]]; then
    echo "Output file '$OUTFILE' exists. Overwrite? (y/n)"
    read OVERWRITE
    if [ $OVERWRITE == "yes" ] || [ $OVERWRITE == "y" ]; then
       echo "Will overwrite."
       rm "$OUTFILE"
    else
       echo "Nothing to do. Exiting."
       exit 0
    fi
fi

if [[ ! -f "$2" ]]; then
   echo "Donor file not found. Using an empty one."
   dd if=/dev/zero bs=1 count=147280 status=none > "$DONOR"

   # Empty slots
   for (( a=0; a<5; a++ ))
   do
      printf "$(printf '\\x%02X' 0x07)" | dd of="$DONOR" bs=1 seek=$(( 112 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x74)" | dd of="$DONOR" bs=1 seek=$(( 113 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0xB9)" | dd of="$DONOR" bs=1 seek=$(( 114 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0xD1)" | dd of="$DONOR" bs=1 seek=$(( 115 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=$(( 4737 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=$(( 5905 + $a * 21008 )) count=1 conv=notrunc status=none
   done
   # Quicksave
   printf "$(printf '\\x%02X' 0x1E)" | dd of="$DONOR" bs=1 seek=105152 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x5F)" | dd of="$DONOR" bs=1 seek=105153 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x0B)" | dd of="$DONOR" bs=1 seek=105154 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x34)" | dd of="$DONOR" bs=1 seek=105155 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=114385 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=115553 count=1 conv=notrunc status=none
else
   DONOR="$2"
   # Detect ff6mobile
   FF6MOBILE=`"${SCRIPTDIR}/util/ff6detect.sh" "$DONOR"`
   if [ "${FF6MOBILE}" != "ff6mobile" ]; then
      echo "$0: Donor file format failure." 1>&2
      exit 3
   fi
fi

echo "Internal checks completed. Starting import."

echo "Importing data slot"

# Header
echo "46494E414C2046414E54415359205649" | xxd -p -r >> "$OUTFILE"
echo "20202020206D6F62696C652020202020" | xxd -p -r >> "$OUTFILE"
echo "02020000000000000000000000000000" | xxd -p -r >> "$OUTFILE"
echo "0000000001000000000000001BE41BE41BE41BE4" | xxd -p -r >> "$OUTFILE"

# Action counter? from donor
tail -c +69 "$DONOR" | head -c 4 >> "$OUTFILE"

dd if=/dev/zero bs=1 count=56 status=none >> "$OUTFILE"

# Data slot
head -c +2560 "$1" >> "$OUTFILE"

# Data slot extansion
echo "Importing gba data extension"
tail -c +8193 "$1" | head -c 1024 >> "$OUTFILE"

# Bestiary
echo "Importing bestiary"
tail -c +11265 "$1" | head -c 1024 >> "$OUTFILE"

# 4 bytes from donor
tail -c +4737 "$DONOR" | head -c 4 >> "$OUTFILE"

# Names
for (( i=0; i<16; i++ ))
do
   NAMEHEX=`xxd -p -l 6 -s $(( 2 + $i * 37 )) -u "$1" | tr -d " \n"`
   NAME=`"${SCRIPTDIR}/util/ff6name.sh" "$NAMEHEX" | tr -d " "`
   echo "Import name: $NAME"
   ZEROS=$(( 6 - ${#NAME} ))
   echo -n "${NAME}" >> "$OUTFILE"
   dd if=/dev/zero bs=1 count=$(( 21 + $ZEROS )) status=none >> "$OUTFILE"
done

# Rest of the data
tail -c +5173 "$DONOR" >> "$OUTFILE"

# Remove possibly created empty donor file.
if [[ -f "/tmp/empty_mob$$.sav" ]]; then
   rm "/tmp/empty_mob$$.sav"
fi

echo "Import completed; '$OUTFILE' created."
echo "Running checksum updater."

"$SCRIPTDIR/util/ff6mob_checksumupdater.sh" "$OUTFILE"

