#!/bin/bash

AUTHOR="Janne Hellas"
VERSION="0.4"
LICENSE="Public Domain or equivalent."
LINFO="This software is provided \"as-is\" without any warranties or liabilities."

SCRIPTDIR=$(dirname "${BASH_SOURCE}")

# Test environment
BC=`which bc`
if [[ "${BC}" == "" ]]; then
   echo "$0: Environmental failure. Install bc." 1>&2
   exit 1
fi

if [[ $# -eq 0 ]]; then
   echo "FF3/6 SNES to PC/Mobile save converter v$VERSION"
   echo ""
   echo "Usage: ./ff6mob_snessave_import.sh [SNES Save file] [Donor PC/Mobile save file]"
   echo ""
   echo "The utility will create a file called 'data_import' from the input files."
   echo "Relevant slot 1 data blocks are replaced by the input SNES save file's slot 1 data."
   echo "Rest of the data is augmented from the donor file. If no donor is specified, an empty one will be utilized."
   echo "Nothing is written to the input files."
   echo ""
   echo "License: $LICENSE"
   echo "$LINFO"
   exit 0
fi

DONOR="/tmp/empty_mob$$.sav"
# handle non-option arguments
if [[ $# -gt 2 ]]; then
    echo "$0: Max two arguments are required." 1>&2
    exit 1
fi

if [[ ! -f "$1" ]]; then
   echo "$0: SNES save file not found." 1>&2
   exit 1
fi

# Detect ff3snes
FF3SNES=`"${SCRIPTDIR}/util/ff6detect.sh" "$1"`
if [ "${FF3SNES}" != "ff3snes" ]; then
    echo "$0: ff3snes save file not detected." 1>&2
    exit 2
fi

# Static parameters
SLOT=1
OUTFILE="data_import"

if [[ "$1" == "$OUTFILE" ]] || [[ "$2" == "$OUTFILE" ]]; then
   echo "$0: Input file cannot be named the same as output." 1>&2
   exit 3
fi

if [[ -f "$OUTFILE" ]]; then
    echo "Output file '$OUTFILE' exists. Overwrite? (y/n)"
    read OVERWRITE
    if [ $OVERWRITE == "yes" ] || [ $OVERWRITE == "y" ]; then
       echo "Will overwrite."
       rm "$OUTFILE"
    else
       echo "Nothing to do. Exiting."
       exit 0
    fi
fi

if [[ ! -f "$2" ]]; then
   echo "Donor file not found. Using an empty one."
   dd if=/dev/zero bs=1 count=147280 status=none > "$DONOR"

   # Empty slots
   for (( a=0; a<5; a++ ))
   do
      printf "$(printf '\\x%02X' 0x07)" | dd of="$DONOR" bs=1 seek=$(( 112 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x74)" | dd of="$DONOR" bs=1 seek=$(( 113 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0xB9)" | dd of="$DONOR" bs=1 seek=$(( 114 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0xD1)" | dd of="$DONOR" bs=1 seek=$(( 115 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=$(( 4737 + $a * 21008 )) count=1 conv=notrunc status=none
      printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=$(( 5905 + $a * 21008 )) count=1 conv=notrunc status=none
   done
   # Quicksave
   printf "$(printf '\\x%02X' 0x1E)" | dd of="$DONOR" bs=1 seek=105152 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x5F)" | dd of="$DONOR" bs=1 seek=105153 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x0B)" | dd of="$DONOR" bs=1 seek=105154 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x34)" | dd of="$DONOR" bs=1 seek=105155 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=114385 count=1 conv=notrunc status=none
   printf "$(printf '\\x%02X' 0x01)" | dd of="$DONOR" bs=1 seek=115553 count=1 conv=notrunc status=none
else
   DONOR="$2"
   # Detect ff6mobile
   FF6MOBILE=`"${SCRIPTDIR}/util/ff6detect.sh" "$DONOR"`
   if [ "${FF6MOBILE}" != "ff6mobile" ]; then
      echo "$0: ff6mobile save not detected on donor file." 1>&2
      exit 2
   fi
fi

echo "Internal checks completed. Starting import."

echo "Importing data slot"

# Header
echo "46494E414C2046414E54415359205649" | xxd -p -r >> "$OUTFILE"
echo "20202020206D6F62696C652020202020" | xxd -p -r >> "$OUTFILE"
echo "02020000000000000000000000000000" | xxd -p -r >> "$OUTFILE"
echo "0000000001000000000000001BE41BE41BE41BE4" | xxd -p -r >> "$OUTFILE"

# Action counter? from donor
tail -c +69 "$DONOR" | head -c 4 >> "$OUTFILE"

dd if=/dev/zero bs=1 count=56 status=none >> "$OUTFILE"

# Data slot

# Item data
ITEM_OFFSET=617
ITEMADV_OFFSET=8208
# Item data slot size for items in advance version
ITEMSLOTADV_SIZE_OFFSET=1024

# Inventory size, bytes
ITEMS_SIZE=256
ITEMSADV_SIZE=288

ITEMHEX=`xxd -p -l $ITEMS_SIZE -s "$ITEM_OFFSET" -u "$1" | tr -d " \n"`
NBHEX=`xxd -p -l $ITEMS_SIZE -s "$(( $ITEM_OFFSET + $ITEMS_SIZE ))" -u "$1" | tr -d " \n"`

#echo $ITEMHEX
#echo $NBHEX

ESPERHEX=`xxd -p -l 4 -s 1129 -u "$1" | tr -d " \n"`
ACTIVEGROUPE=`xxd -p -l 1 -s 1133 -u "$1" | tr -d " \n"`
MORPHBAR=`xxd -p -l 1 -s 1782 -u "$1" | tr -d " \n"`
BUSHIDO=`xxd -p -l 1 -s 1783 -u "$1" | tr -d " \n"`

#echo $ESPERHEX
#echo $ACTIVEGROUPE

head -c $ITEM_OFFSET "$1" >> "$OUTFILE"

# Magic, 54 bytes per character
MAGIC_OFFSET=1134
# 54 spells
MAGICDATASLOT_SIZE=54
# 12 slots for 12 characters
MAGICSLOTS=12

# For Advance version
MAGICADV_OFFSET=617
EXTRA_OFFSET=7
MAGICDATASLOTADV_SIZE=57

MAGICHEX=`xxd -p -l $(( $MAGICDATASLOT_SIZE * $MAGICSLOTS )) -s $MAGIC_OFFSET -u "$1" | tr -d " \n"`

#echo $MAGICHEX

# Patch Magic data

for (( i=0; i<${MAGICSLOTS}; i++ ))
do
   echo ${MAGICHEX:$(( $i * $MAGICDATASLOT_SIZE * 2 )):$(( $MAGICDATASLOT_SIZE * 2 ))} | xxd -p -r >> "$OUTFILE"
   # Zero GBA specific spells & extra space
   dd if=/dev/zero bs=1 count=10 status=none >> "$OUTFILE"
done

# More zeros
dd if=/dev/zero bs=1 count=396 status=none >> "$OUTFILE"

# Write active group number
echo ${ACTIVEGROUPE} | xxd -p -r >> "$OUTFILE"

# Write morphbar
echo "${MORPHBAR}" | xxd -p -r >> "$OUTFILE"

# Write bushido
echo "${BUSHIDO}" | xxd -p -r >> "$OUTFILE"

# Bushido name marks
echo "D3AA9EEF90FFD3AA9EEF9AFFD3AA9EEFA0FFD3AA9EEFD5FFD3AA9EEFE9FFD3AA9EEF9DFFD3AA9EEFECFFD3AA9EEFC2FF" | xxd -p -r >> "$OUTFILE"

# Copy until empty 0x55 blocks and zero them
tail -c +1833 "$1" | head -c 245 >> "$OUTFILE"
dd if=/dev/zero bs=1 count=35 status=none >> "$OUTFILE"

tail -c +2113 "$1" | head -c 48 >> "$OUTFILE"
dd if=/dev/zero bs=1 count=16 status=none >> "$OUTFILE"

tail -c +2177 "$1" | head -c 322 >> "$OUTFILE"

tail -c +2499 "$1" | head -c 16 | sed 's/\x55\x55/\x00\x00/g' >> "$OUTFILE"

# Parent facing direction
tail -c +2515 "$1" | head -c 1 >> "$OUTFILE"

# Saved XY
for (( i=0; i<32; i=i+2 ))
do
   tail -c +$(( 2516 + $i * 2 )) "$1" | head -c 2 | sed 's/\x55\x55/\x00\x00/g' >> "$OUTFILE"
done

# Z-levels
tail -c +2548 "$1" | head -c 4 >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=9 status=none >> "$OUTFILE"

# Reset save counter, just in case
printf "$(printf '\\x%02X' 0x00)" | dd of="$OUTFILE" bs=1 seek=2119 count=1 conv=notrunc status=none
printf "$(printf '\\x%02X' 0x00)" | dd of="$OUTFILE" bs=1 seek=2120 count=1 conv=notrunc status=none

# Data extension

# Espers
echo $ESPERHEX | xxd -p -r >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=12 status=none >> "$OUTFILE"

# Items
echo "$ITEMHEX" | xxd -p -r >> "$OUTFILE"

# Empty GBA specific inventory
echo "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF" | xxd -p -r >> "$OUTFILE"

# Item counts
echo "$NBHEX" | xxd -p -r >> "$OUTFILE"

# Empty GBA specific counts + GBA eventbits
dd if=/dev/zero bs=1 count=48 status=none >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=412 status=none >> "$OUTFILE"

# Extension slot mark + space for checksum
echo "474D0000" | xxd -p -r >> "$OUTFILE"

# Test whether to add Stone Table to inventory
TEST=$(( 0x${ESPERHEX:2:2} & 0x80 ))
if [[ $TEST -eq 128 ]]; then
   echo "Adding Stone Tablet to inventory."
   RAREI=`xxd -p -l 1 -s 2236 -u "$1"`
   NEWVALUE=$(( 0x$RAREI | 0x10 ))
   NEWVALUE2=$(echo "obase=16; ibase=10; $NEWVALUE" | bc)
   #echo "Replace: $RAREI, $NEWVALUE, $NEWVALUE2"
   printf "$(printf '\\x%02X' 0x$NEWVALUE2)" | dd of="$OUTFILE" bs=1 seek=2364 count=1 conv=notrunc status=none
fi

# Take data until names
tail -c +3713 "$DONOR" | head -c 1028 >> "$OUTFILE"

# Names
for (( i=0; i<16; i++ ))
do
   NAMEHEX=`xxd -p -l 6 -s $(( 2 + $i * 37 )) -u "$1" | tr -d " \n"`
   NAME=`"${SCRIPTDIR}/util/ff6name.sh" "$NAMEHEX" | tr -d " "`
   echo "Import name: $NAME"
   ZEROS=$(( 6 - ${#NAME} ))
   echo -n "${NAME}" >> "$OUTFILE"
   dd if=/dev/zero bs=1 count=$(( 21 + $ZEROS )) status=none >> "$OUTFILE"
done

# Rest of the data
tail -c +5173 "$DONOR" >> "$OUTFILE"

# Remove possibly created empty donor file.
if [[ -f "/tmp/empty_mob$$.sav" ]]; then
   rm "/tmp/empty_mob$$.sav"
fi

echo "Import completed; '$OUTFILE' created."
echo "Running checksum updater."

"$SCRIPTDIR/util/ff6mob_checksumupdater.sh" "$OUTFILE"

